"""Module base for the Python reference specification of SOMA.

Types will be defined in their own modules and then imported here for a single
unified namespace.
"""

from somacore import _version
from somacore import base
from somacore import data
from somacore import options

__version__ = _version.version
__version_tuple__ = _version.version_tuple

SOMAObject = base.SOMAObject
Collection = base.Collection

DataFrame = data.DataFrame
NDArray = data.NDArray
DenseNDArray = data.DenseNDArray
SparseNDArray = data.SparseNDArray

IOfN = options.IOfN
BatchSize = options.BatchSize

__all__ = (
    "SOMAObject",
    "Collection",
    "DataFrame",
    "NDArray",
    "DenseNDArray",
    "SparseNDArray",
    "IOfN",
    "BatchSize",
)



"""Internal version information to indirect `_generated_version.py`.

setuptools imports this module to find the version for the package at *build
time*. We use this indirect module so that if the user has not run
``write-version-file`` (e.g. they just did a fresh checkout), ``pip install``
will still work. If we used ``setuptools.dynamic.version.attr`` and pointed it
at `somacore._generated_version` directly, installing from a fresh checkout
would fail.
"""

import pathlib
from typing import Any, Dict, Tuple, Union


def _read_version() -> Tuple[str, Tuple[Union[str, int], ...]]:
    version_contents: Dict[str, Any] = dict(
        version="0.0.0.dev+local-checkout",
        version_tuple=(0, 0, 0, "dev", "local-checkout"),
    )

    # We do this the hard way since neither
    #     from somacore import _generated_version
    # nor
    #     from . import _generated_version
    # will work in the build environment.
    gen_file = pathlib.Path(__file__).parent / "_generated_version.py"
    try:
        exec(gen_file.read_text(), version_contents)
    except Exception:
        # The generated file does not exist or is not valid.
        # Ignore it.
        pass
    return version_contents["version"], version_contents["version_tuple"]


version, version_tuple = _read_version()



"""Definitions of the most fundamental types used by the SOMA project.

SOMA users should ordinarily not need to import this module directly; relevant
members will be exported to the ``somacore`` namespace.
"""

import abc
from typing import Any, MutableMapping, TypeVar

from typing_extensions import LiteralString


class SOMAObject(metaclass=abc.ABCMeta):
    """A sentinel interface indicating that this type is a SOMA object."""

    __slots__ = ("__weakref__",)

    @property
    def context(self) -> Any:
        """A value storing implementation-specific configuration information.

        This contains long-lived (i.e., not call-specific) information that is
        used by the SOMA implementation to access storage. This may include
        things like credentials, endpoint locations, or database connections.

        End users should treat this as an opaque value. While it may be passed
        from an existing SOMA object to be used in the creation of a new SOMA
        object, it should not be inspected.
        """
        return None

    @property
    @abc.abstractmethod
    def metadata(self) -> MutableMapping[str, Any]:
        """The metadata of this SOMA object.

        The returned value directly references the stored metadata; reads from
        and writes to it (provided the object is opened) are reflected in
        storage.
        """
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def soma_type(self) -> LiteralString:
        """A string describing the SOMA type of this object."""
        raise NotImplementedError()


_ST = TypeVar("_ST", bound=SOMAObject)
"""Generic type variable for any SOMA object."""


class Collection(SOMAObject, MutableMapping[str, _ST]):
    """A generic string-keyed collection of :class:`SOMAObject`s.

    The generic type specifies what type the Collection may contain. At its
    most generic, a Collection may contain any SOMA object, but a subclass
    may specify that it is a Collection of a specific type of SOMA object.
    """

    __slots__ = ()

    # This is implemented as a property and not a literal so that it can be
    # overridden with `Final` members in Collection specializations.
    @property
    def soma_type(self) -> LiteralString:
        return "SOMACollection"



"""Implementations of the composed SOMA data types."""

from typing_extensions import Final

from somacore import _wrap
from somacore import base
from somacore import data


class Measurement(_wrap.CollectionProxy):
    """A set of annotated variables and values for a single measurement."""

    __slots__ = ()

    var = _wrap.item(data.DataFrame)
    """Primary annotations on the variable axis for vars on this meansurement.

    This annotates _columns_ of the ``X`` arrays. The contents of the
    ``soma_joinid`` pseudo-column define the variable index domain (``varid``)
    All variables for this measurement _must_ be defined in this dataframe.
    """

    X = _wrap.item(base.Collection[data.NDArray])
    """A collection of matrices containing feature values.

    Each matrix is indexed by ``[obsid, varid]``. Sparse and dense 2D arrays may
    both be used in any combination in ``X``.
    """

    obsm = _wrap.item(base.Collection[data.DenseNDArray])
    """Matrices containing annotations of each ``obs`` row.

    This has the same shape as ``obs`` and is indexed with ``obsid``.
    """

    obsp = _wrap.item(base.Collection[data.SparseNDArray])
    """Matrices containg pairwise annotations of each ``obs`` row.

    This is indexed by ``[obsid_1, obsid_2]``.
    """

    varm = _wrap.item(base.Collection[data.DenseNDArray])
    """Matrices containing annotations of each ``var`` row.

    This has the same shape as ``var`` and is indexed with ``varid``.
    """

    varp = _wrap.item(base.Collection[data.SparseNDArray])
    """Matrices containg pairwise annotations of each ``var`` row.

    This is indexed by ``[varid_1, varid_2]``.
    """

    soma_type: Final = "SOMAMeasurement"



"""Definitions of data storage interfaces for SOMA implementations.

SOMA users should ordinarily not need to import this module directly; relevant
members will be exported to the ``somacore`` namespace.
"""

import abc
from typing import Any, Iterator, Optional, Sequence, Tuple, TypeVar, Union

import pyarrow
from typing_extensions import Final

from somacore import base
from somacore import options


class DataFrame(base.SOMAObject, metaclass=abc.ABCMeta):
    """A multi-column table with a user-defined schema."""

    __slots__ = ()

    # Data operations

    @abc.abstractmethod
    def read(
        self,
        ids,  # TODO: Specify type
        column_names: Optional[Sequence[str]] = None,
        *,
        batch_size: options.BatchSize = options.BatchSize(),
        partitions: Optional[options.ReadPartitions] = None,
        result_order: Optional[options.ResultOrder] = None,
        value_filter: Optional[str] = None,
        platform_config: Optional[options.PlatformConfig] = None,
    ) -> "ReadIter[pyarrow.Table]":
        """Reads a user-defined slice of data into Arrow tables.

        TODO: Further per-param documentation.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def write(
        self,
        values: Union[pyarrow.RecordBatch, pyarrow.Table],
        platform_config: Optional[options.PlatformConfig] = None,
    ) -> None:
        """Writes values to the data store.

        TODO: Further per-param documentation.
        """
        raise NotImplementedError()

    # Metadata operations

    @property
    @abc.abstractmethod
    def schema(self) -> pyarrow.Schema:
        """The schema of the data in this dataframe."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def index_column_names(self) -> Tuple[str, ...]:
        """The names of the index (dimension) columns."""
        raise NotImplementedError()

    # Basic operations

    soma_type: Final = "SOMADataFrame"


class NDArray(base.SOMAObject, metaclass=abc.ABCMeta):
    """Common behaviors of N-dimensional arrays of a single primitive type."""

    __slots__ = ()

    # Metadata operations

    @property
    @abc.abstractmethod
    def shape(self) -> Tuple[int, ...]:
        """The length of each dimension of this array."""
        raise NotImplementedError()

    @property
    def ndims(self) -> int:
        """The Number of Dimensions in this array."""
        return len(self.shape)

    @property
    @abc.abstractmethod
    def schema(self) -> pyarrow.Schema:
        """The schema of the data in this array."""
        raise NotImplementedError()

    @property
    @abc.abstractmethod
    def is_sparse(self) -> bool:
        """True if this array is sparse. False if this array is dense."""
        raise NotImplementedError()


class DenseNDArray(NDArray, metaclass=abc.ABCMeta):
    """A N-dimensional array stored densely."""

    __slots__ = ()

    @abc.abstractmethod
    def read(
        self,
        coords: options.ReadCoords,
        *,
        batch_size: options.BatchSize = options.BatchSize(),
        partitions: Optional[options.ReadPartitions] = None,
        result_order: options.ResultOrder = None,  # TODO: Set default
        platform_config: Optional[options.PlatformConfig] = None,
    ) -> pyarrow.Tensor:
        """Reads the specified subarray from this NDArray as a Tensor.

        TODO: Additional per-param documentation.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def write(
        self,
        coords: options.ReadCoords,
        values: pyarrow.Tensor,
        *,
        platform_config: Optional[options.PlatformConfig] = None,
    ) -> None:
        """Writes a Tensor to a subarray of the persistent object.

        TODO: Additional per-param documentation.
        """
        raise NotImplementedError()

    is_sparse: Final = False
    soma_type: Final = "SOMADenseNDArray"


class SparseNDArray(NDArray, metaclass=abc.ABCMeta):
    """A N-dimensional array stored sparsely."""

    __slots__ = ()

    @abc.abstractmethod
    def read(
        self,
        slices: Any,  # TODO: Define this type.
        *,
        batch_size: options.BatchSize = options.BatchSize(),
        partitions: Optional[options.ReadPartitions] = None,
        result_order: options.ResultOrder = None,  # TODO: Set default
        platform_config: Optional[options.PlatformConfig] = None,
    ) -> "SparseRead":
        """Reads a subset of the object in one or more batches.

        Values returned are a :class:`SparseRead` object which can be converted
        to any number of formats::

            some_dense_array.read(...).tables()
            # -> an iterator of Arrow Tables
            some_dense_array.read(...).csrs().all()
            # -> a single flattened sparse CSR matrix

        TODO: Additional per-param documentation.
        """

    @abc.abstractmethod
    def write(
        self,
        values: pyarrow.Tensor,
        *,
        platform_config: Optional[options.PlatformConfig] = None,
    ) -> None:
        """Writes a Tensor to a subarray of the persistent object.

        TODO: Additional per-param documentation.
        """
        raise NotImplementedError()

    @property
    def nnz(self) -> int:
        """The number of values stored in the array, including explicit zeros.

        For dense arrays, this will be the total size of the array.
        """
        raise NotImplementedError()

    is_sparse: Final = True
    soma_type: Final = "SOMASparseNDArray"


#
# Read types
#

_T = TypeVar("_T")


# Sparse reads are returned as an iterable structure:


class ReadIter(Iterator[_T], metaclass=abc.ABCMeta):
    """SparseRead result iterator allowing users to flatten the iteration."""

    # __iter__ is already implemented as `return self` in Iterator.
    # SOMA implementations must implement __next__.

    @abc.abstractmethod
    def concat(self) -> _T:
        """Returns all the requested data in a single operation.

        If some data has already been retrieved using ``next``, this will return
        the rest of the data after that is already returned.
        """
        raise NotImplementedError()


class SparseRead:
    """Intermediate type to choose result format when reading a sparse array.

    A query may not be able to return all of these formats. The concrete result
    may raise a ``NotImplementedError`` or may choose to raise a different
    exception (likely a ``TypeError``) containing more specific information
    about why the given format is not supported.
    """

    def coos(self) -> ReadIter[pyarrow.SparseCOOTensor]:
        raise NotImplementedError()

    def cscs(self) -> ReadIter[pyarrow.SparseCSCMatrix]:
        raise NotImplementedError()

    def csrs(self) -> ReadIter[pyarrow.SparseCSRMatrix]:
        raise NotImplementedError()

    def dense_tensors(self) -> ReadIter[pyarrow.Tensor]:
        raise NotImplementedError()

    def record_batches(self) -> ReadIter[pyarrow.RecordBatch]:
        raise NotImplementedError()

    def tables(self) -> ReadIter[pyarrow.Table]:
        raise NotImplementedError()



from typing import (
    Any,
    Generic,
    ItemsView,
    Iterator,
    KeysView,
    MutableMapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    ValuesView,
    overload,
)

import attrs
from typing_extensions import Final

from somacore import base

_ST = TypeVar("_ST", bound=base.SOMAObject)
_T = TypeVar("_T")

_SENTINEL = object()


class CollectionProxy(base.Collection[_ST]):
    """Base class to forward SOMA collection methods to a "real" SOMA object.

    This is intended for use as a mixin for SOMA objects that decorate standard
    collection types with behaviors, e.g. the Experiment or Measurement classes.
    It allows the separation of the behaviors (implemented in the decorator)
    from the backend implementation (implemented in the wrapped object).
    """

    __slots__ = ("_backing",)

    def __init__(self, backing: base.Collection[_ST]):
        """Creates a new CollectionProxy backed by the given collection."""
        self._backing: Final = backing
        """The object that actually provides the indexing for this object."""

    # SOMA methods

    @property
    def metadata(self) -> MutableMapping[str, str]:
        return self._backing.metadata

    def unwrap(self) -> base.Collection[_ST]:
        """Unwrap gets the actual collection backed by this proxy."""
        inst: base.Collection[_ST] = self
        while isinstance(inst, CollectionProxy):
            inst = inst._backing
        return inst

    # Sized

    def __len__(self) -> int:
        return len(self._backing)

    # Iterable

    def __iter__(self) -> Iterator[str]:
        return iter(self._backing)

    # Container

    def __contains__(self, key: Any) -> bool:
        return key in self._backing

    # Mapping

    def __getitem__(self, key: str) -> _ST:
        return self._backing[key]

    @overload
    def get(self, key: str) -> Optional[_ST]:
        ...

    @overload
    def get(self, key: str, default: _T) -> Union[_ST, _T]:
        ...

    def get(
        self, key: str, default: Union[_ST, _T, None] = None
    ) -> Union[_ST, _T, None]:
        return self._backing.get(key, default)

    def keys(self) -> KeysView[str]:
        return self._backing.keys()

    def items(self) -> ItemsView[str, _ST]:
        return self._backing.items()

    def values(self) -> ValuesView[_ST]:
        return self._backing.values()

    # __reversed__ is not supported.

    # MutableMapping

    def __setitem__(self, key: str, value: _ST) -> None:
        self._backing[key] = value

    def __delitem__(self, key: str) -> None:
        del self._backing[key]

    def pop(
        self,
        key: str,
        value: Union[_T, _ST] = _SENTINEL,  # type: ignore[assignment]
    ) -> Union[_T, _ST]:
        if value is _SENTINEL:
            return self._backing.pop(key)
        return self._backing.pop(key, value)

    def popitem(self) -> Tuple[str, _ST]:
        return self._backing.popitem()

    def clear(self) -> None:
        self._backing.clear()

    def update(self, __other=(), **kwds) -> None:
        return self._backing.update(__other, **kwds)

    def setdefault(
        self,
        key: str,
        default: _ST = None,  # type: ignore[assignment]
    ) -> _ST:
        return self._backing.setdefault(key, default)  # type: ignore[arg-type]


@attrs.define()
class item(Generic[_T]):
    """Descriptor to transform property access into indexing.

    This descriptor works with :class:`CollectionProxy` to allow for simple
    specification of properties that reflect the members of the collection::

        class WrapImpl(CollectionProxy):

            first = item(str)
            second = item(int, "2nd")

        inst = WrapImpl(some_collection)

        # This is equivalent to getting some_collection["first"]
        inst.first

        # This is equivalent to setting some_collection["2nd"]
        inst.second = 500
    """

    typ: Optional[Type[_T]] = None
    """The type we expect to return from this field."""

    item_name: Optional[str] = None
    """The name of the item we are getting (``x._backing["whatever"]``).

    This uses the name of the field by default but can be manually overridden.
    """

    field_name: str = attrs.field(default="<unknown>", init=False)
    """The name of this field (``x.whatever``). Set automatically."""

    def __set_name__(self, owner: Type[CollectionProxy], name: str) -> None:
        del owner  # unused
        self.field_name = name
        if self.item_name is None:
            self.item_name = name

    @overload
    def __get__(self, inst: None, owner: Type[CollectionProxy]) -> "item[_T]":
        ...

    @overload
    def __get__(self, inst: CollectionProxy, owner: Type[CollectionProxy]) -> _T:
        ...

    def __get__(
        self, inst: Optional[CollectionProxy], owner: Type[CollectionProxy]
    ) -> Union["item", _T]:
        del owner  # unused
        if not inst:
            return self
        assert self.item_name is not None
        try:
            # TODO: Type-check params/returns?
            return inst[self.item_name]
        except KeyError as ke:
            raise AttributeError(
                f"{_typename(inst)!r} object has no attribute {self.field_name!r}"
            ) from ke

    def __set__(self, inst: CollectionProxy, value: _T) -> None:
        assert self.item_name is not None
        # Pretend it's a MutableMapping for the type-checker.
        # If it fails that's OK; we need to raise anyway.
        try:
            inst[self.item_name] = value
        except KeyError as ke:
            raise AttributeError(
                f"{_typename(inst)!r} does not support assigning"
                f" to item {self.item_name!r}"
            ) from ke

    def __delete__(self, inst: CollectionProxy) -> None:
        assert self.item_name is not None
        try:
            del inst[self.item_name]
        except KeyError as ke:
            raise AttributeError(
                f"{_typename(inst)} does not support deleting {self.item_name!r}"
            ) from ke


def _typename(x: object) -> str:
    return type(x).__name__
